"""
Socket Pool used by the client to manage connections to the server
"""
import heapq
import time
from io import BufferedRWPair
from typing import Dict, List, Tuple, Any, Callable, Union
from gevent import socket
from gevent.thread import get_ident


class SocketPool:
    """
    Manages connections to the kvault server
    """

    # pylint: disable-next=missing-function-docstring
    def __init__(self, host: str, port, max_age: int = 60):
        self.host = host
        self.port = port
        self.max_age = max_age
        self.free: List[Tuple[float, BufferedRWPair]] = []
        self.in_use: Dict[Union[int, BufferedRWPair], BufferedRWPair] = {}
        self._tid: Callable[[Any], int] = get_ident

    # pylint: disable-next=missing-function-docstring
    def checkout(self):
        """Establishes a socket connection"""
        now = time.time()
        tid = self._tid()
        if tid in self.in_use:
            sock = self.in_use[tid]
            if sock.closed:
                del self.in_use[sock]
            else:
                return self.in_use[sock]

        while self.free:
            timestamp, sock = heapq.heappop(self.free)
            if timestamp < now - self.max_age:
                try:
                    sock.close()
                except OSError:
                    pass
            else:
                self.in_use[tid] = sock
                return sock

        sock = self.create_socket_file()
        self.in_use[tid] = sock
        return sock

    def checkin(self):
        """Checks if there is an in use socket connection and adds it to the heap returning a boolean value. True
        indicates that the socket is in use, false indicates it is not"""
        tid = self._tid()
        if tid in self.in_use:
            sock = self.in_use.pop(tid)
            if not sock.closed:
                heapq.heappush(self.free, (time.time(), sock))
            return True
        return False

    def close(self):
        """Closes connection of the socket"""
        tid = self._tid()
        sock = self.in_use.pop(tid, None)
        if sock:
            try:
                sock.close()
            except OSError:
                pass
            return True
        return False

    def create_socket_file(self):
        """Connects to socket on the AF_INET family with the SOCK_STREAM socket kind and returns a BufferedRWPair"""
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        conn.connect((self.host, self.port))
        return conn.makefile("rwb")



"""
Entry point to server
"""
import argparse
import importlib
from .queue_server import QueueServer
from .infra.logger import logger


def get_args_parser():
    """
    Options parser. parses the options that will be used on the command line and returns a parser
    """
    parser = argparse.ArgumentParser(prog="KVault", description="Key Value Store")

    parser.add_argument(
        "-d", "--debug", action="store_true", dest="debug", help="Log debug messages."
    )
    parser.add_argument(
        "-e",
        "--errors",
        action="store_true",
        dest="error",
        help="Log error messages only.",
    )
    parser.add_argument(
        "-H", "--host", default="127.0.0.1", dest="host", help="Host to listen on."
    )
    parser.add_argument(
        "-m",
        "--max-clients",
        default=1024,
        dest="max_clients",
        help="Maximum number of clients.",
        type=int,
    )
    parser.add_argument(
        "-p", "--port", default=31337, dest="port", help="Port to listen on.", type=int
    )
    parser.add_argument("-l", "--log-file", dest="log_file", help="Log file.")
    parser.add_argument(
        "-x",
        "--extension",
        action="append",
        dest="extensions",
        help="Import path for Python extension module(s).",
    )
    return parser


def load_extensions(server, extensions):
    """
    Loads extensions that will be added and initialized with the server.
    :param server: Server instance
    :param extensions: list of extensions to load
    :return:
    """
    for extension in extensions:
        try:
            module = importlib.import_module(extension)
        except ImportError:
            logger.exception(f"Could not import extension {extension}")
        else:
            try:
                initialize = getattr(module, "initialize")
            except AttributeError:
                logger.exception(
                    f'Could not find "initialize" function in extension {extension}'
                )
                raise

            initialize(server)
            logger.info(f"Loaded {extension} extension")


if __name__ == "__main__":
    args = get_args_parser().parse_args()

    from gevent import monkey

    monkey.patch_all()

    # configure_logger(options)
    queue_server = QueueServer(
        host=args.host, port=args.port, max_clients=args.max_clients
    )
    load_extensions(queue_server, args.extensions or ())
    print("\x1b[32m  .--.")
    print(
        # pylint: disable-next=consider-using-f-string
        " /( \x1b[34m@\x1b[33m >\x1b[32m    ,-.  "
        "\x1b[1;32mKVault "
        "\x1b[1;33m%s:%s\x1b[32m" % (args.host, args.port)
    )
    print("/ ' .'--._/  /")
    print(":   ,    , .'")
    print("'. (___.'_/")
    print(" \x1b[33m((\x1b[32m-\x1b[33m((\x1b[32m-''\x1b[0m")
    try:
        queue_server.run()
    except KeyboardInterrupt:
        print("\x1b[1;31mshutting down\x1b[0m")



"""
Protocol Handler is based on the Redis Wire protocol. The server uses the protocol handler to unpack client requests
and serialize server responses back to the client.
"""
import datetime
from typing import Union, Optional, List, Dict, Set, Any, Callable
import json
from io import BytesIO
from collections import deque
from .exceptions import Error
from .types import unicode
from .utils import encode
from .utils.mixins import MetaUtils
from .infra.logger import logger


class ProtocolHandler(MetaUtils):
    """
    ProtocolHandler is based on Redis Wire protocol which uses a request/response communication pattern with clients.
    Responses from the server will use the first byte to indicate data-type, followed by the data, terminated by a
    carriage-return/line

    Client sends requests as an array of bulk strings.

    Server replies, indicating response type using the first byte:

    Data-type | Prefix |  Structure | Example

    simple string | "+" | +{simple string}\r\n | +this is a string\r\n
    Simple strings: "+string content\r\n"  <-- cannot contain newlines

    Error | "-" | -{error message}\r\n | -Err unknown command "OOPS"\r\n
    Error: "-Error message\r\n"

    Integer/float | ":" | :{number}\r\n | :12\r\n
    Integers: ":1337\r\n"

    Binary(or bulk string | "$" | ${number of bytes}\r\n{data}\r\n | $6\r\nfoobar/r/n
    Bulk String: "$number of bytes\r\nstring data\r\n"

    Unicode string(or bulk unicode string) | "^" | ^\r\n{data} | ^\r\nfoobar
    Bulk unicode string (encoded as UTF-8): "^number of bytes\r\ndata\r\n"

    Json string(uses bulk string rules) | '@' | @number of bytes\r\nJSON string\r\n |

    Array | "*" | "*number of elements\r\n...elements..." |
    *3\r\n+a simple string element\r\n:12345\r\n$7\r\ntesting\r\n
    Empty array: "*0\r\n"

    Dictionary | "%" | %{number of keys}\r\n{0 or more of above}\r\n |
    %3\r\n+key1\r\n+value1\r\n+key2\r\n*2\r\n+value2-0\r\n+value2-1\r\n:3\r\n$7\r\ntesting\r\n
    Dictionary: "%number of elements\r\n...key0...value0...key1...value1..\r\n"

    Set | "&" | "&number of elements\r\n...elements..."

    Other types:
    NULL: "$-1\r\n"
    Empty string: "$0\r\n\r\n"
    """

    # pylint: disable-next=missing-function-docstring
    def __init__(self):
        self.handlers: Dict[bytes, Callable] = {
            b"+": self.handle_simple_string,
            b"-": self.handle_error,
            b":": self.handle_integer,
            b"$": self.handle_string,
            b"^": self.handle_unicode,
            b"@": self.handle_json,
            b"*": self.handle_array,
            b"%": self.handle_dict,
            b"&": self.handle_set,
        }

    def handle_request(self, socket_file) -> bytes:
        """
        Parse a request from the client into its components parts
        :param socket_file: Socket file
        :return: bytes that are sent to the server
        """
        first_byte = socket_file.read(1)
        if not first_byte:
            logger.error(
                f"[{self.name}] failed to handle request, missing first byte {first_byte}"
            )
            raise EOFError()

        handler = self.handlers.get(first_byte, None)
        if handler:
            return handler(socket_file)
        logger.error(
            f"{self.name}> failed to handle request, missing value for key {first_byte}"
        )
        rest = socket_file.readline().rstrip(b"\r\n")
        return first_byte + rest

    def write_response(self, socket_file, data: Any):
        """
        Serialize the response data and send it to the client
        :param socket_file:
        :param data: Data to respond
        """
        buf = BytesIO()
        self._write(buf, data)
        buf.seek(0)
        socket_file.write(buf.getvalue())
        socket_file.flush()

    # pylint: disable-next=too-many-branches
    def _write(self, buf: BytesIO, data: Any):
        """
        Handle serialization of responses on a buffer to the client. This is used by the server when responding to the
        client
        :param buf: Buffer to write responses to
        :param data: Data to respond with
        """
        if isinstance(data, bytes):
            buf.write(b"$%d\r\n%s\r\n" % (len(data), data))
        elif isinstance(data, unicode):
            bdata = data.encode("utf-8")
            buf.write(b"^%d\r\n%s\r\n" % (len(bdata), bdata))
        elif data is True or data is False:
            buf.write(b":%d\r\n" % (1 if data else 0))
        elif isinstance(data, (int, float)):
            buf.write(b":%d\r\n" % data)
        elif isinstance(data, Error):
            buf.write(b"-%s\r\n" % encode(data.message))
        elif isinstance(data, (list, tuple, deque)):
            buf.write(b"*%d\r\n" % len(data))
            for item in data:
                self._write(buf, item)
        elif isinstance(data, dict):
            buf.write(b"%%%d\r\n" % len(data))
            for key in data:
                self._write(buf, key)
                self._write(buf, data[key])
        elif isinstance(data, set):
            buf.write(b"&%d\r\n" % len(data))
            for item in data:
                self._write(buf, item)
        elif data is None:
            buf.write(b"$-1\r\n")
        elif isinstance(data, datetime.datetime):
            self._write(buf, str(data))

    def handle_simple_string(self, socket_file) -> bytes:
        """
        Handles simple string serialization sent over the wire.
        Format of the data looks like this +{simple string}\r\n
        An example +this is a string\r\n
        :param socket_file: file like object to read data.
        :return: data with the carriage-return/line stripped
        """
        return socket_file.readline().rstrip(b"\r\n")

    def handle_error(self, socket_file) -> Error:
        """
        Handles error serialization sent over the wire.
        Format of the data looks like this -{error message}\r\n
        An example -Err unknown command "OOPS"\r\n
        :param socket_file: file like object to read data.
        :return: data with the carriage-return/line stripped
        """
        return Error(socket_file.readline().rstrip(b"\r\n"))

    def handle_integer(self, socket_file) -> Union[float, int]:
        """
        Handles integer serialization/deserialization sent over the wire.
        Format of the data looks like this :{number}\r\n
        An example :12\r\n
        :param socket_file: file like object to read data.
        :return: data with the carriage-return/line stripped
        """
        number = socket_file.readline().rstrip(b"\r\n")
        if b"." in number:
            return float(number)
        return int(number)

    def handle_string(self, socket_file) -> Optional[bytes]:
        """
        Handles string serialization/deserialization sent over the wire.
        Format of the data looks like this ${number of bytes}\r\n{data}\r\n
        An example $6\r\nfoobar/r/n
        :param socket_file: file like object to read data.
        :return: data with the carriage-return/line stripped or None if the length is less than 0, which is a special
        case for Nulls(None)
        """
        # read the length ($<length>\r\n)
        length = int(socket_file.readline().rstrip(b"\r\n"))
        if length == -1:
            # special case for NULLs
            return None
        # include the trailing \r\n in count
        length += 2
        return socket_file.read(length)[:-2]

    def handle_unicode(self, socket_file) -> Optional[str]:
        """
        Handles unicode characters which in this case calls on the handle_string method to handle the string.
        The format looks the same
        :param socket_file:: File like object to read data
        :return: Optional string
        """
        string_ = self.handle_string(socket_file=socket_file)
        if string_:
            return string_.decode("utf-8")
        return None

    def handle_json(self, socket_file) -> Any:
        """
        Handles JSON string (uses bulk string rules). This uses the handler handle_string
        :param socket_file: File like object to read data
        :return: Python object
        """
        return json.loads(self.handle_string(socket_file=socket_file))

    def handle_array(self, socket_file) -> List[Any]:
        """
        Handles array deserialization. Format that is expected has this format "*number of elements\r\n...elements..."
        where the * prefix is expected. For example: *3\r\n+a simple string element\r\n:12345\r\n$7\r\ntesting\r\n
        :param socket_file: File like object
        :return: list of serialized objects
        """
        num_elements = int(socket_file.readline().rstrip(b"\r\n"))
        return [
            self.handle_request(socket_file=socket_file) for _ in range(num_elements)
        ]

    def handle_dict(self, socket_file) -> Dict[Any, Any]:
        """
        Handles serialization/deserialization of dictionary type. The format looks like
        %{number of keys}\r\n{0 or more of above}\r\n

        where the % prefix is expected.

        For example: %3\r\n+key1\r\n+value1\r\n+key2\r\n*2\r\n+value2-0\r\n+value2-1\r\n:3\r\n$7\r\ntesting\r\n
        :param socket_file: File like object
        :return: Dictionary or key value pairs
        """
        num_items = int(socket_file.readline().rstrip(b"\r\n"))
        elements = [
            self.handle_request(socket_file=socket_file) for _ in range(num_items * 2)
        ]
        return dict(zip(elements[::2], elements[1::2]))

    def handle_set(self, socket_file) -> Set:
        """
        Handle set serialization/deserialization of set objects. This uses the handle_array function and casts this to
        a set
        :param socket_file: File like object
        :return: set data structure
        """
        return set(self.handle_array(socket_file=socket_file))



"""
Mixin classes that provide extra functionality to classes
"""
from typing import Dict, Any
import time
from collections import deque
from ..exceptions import CommandError
from ..types import Value, QUEUE, HASH, SET, KV


# pylint: disable-next=too-few-public-methods
class MetaUtils:
    """
    Class with helper properties and utilities
    """

    @property
    def name(self):
        """Returns the class name"""
        return self.__class__.__name__


class Guards:
    """
    Contains validity checks for the data types and expiry time of commands
    """

    def __init__(self, kv_store: Dict, expiry_map: Dict[Any, float]):
        self._kv = kv_store
        self._expiry_map = expiry_map

    def check_expired(self, key, timestamp=None) -> bool:
        """
        Checks if a key has expired
        :param key: Key
        :param timestamp: Timestamp, defaulted to None and will use current time
        :return: boolean
        """
        _timestamp = timestamp or time.time()
        return key in self._expiry_map and _timestamp > self._expiry_map[key]

    def check_datatype(self, data_type, key, set_missing=True, subtype=None):
        """
        Checks the data type of the supplied key
        :param data_type: Data type to check
        :param key: Key to check
        :param set_missing: Whether to set the value if the key is missing
        :param subtype: subtype to check
        :return: None
        :raises CommandError if the operation is against a wrong key type of wrong value
        """
        if key in self._kv and self.check_expired(key):
            del self._kv[key]

        if key in self._kv:
            value = self._kv[key]
            if value.data_type != data_type:
                raise CommandError(
                    f"Operation against wrong key type. Key type {value.data_type}. data type: {data_type}"
                )
            if subtype is not None and not isinstance(value.value, subtype):
                raise CommandError(
                    f"Operation against wrong value type. Value: {value.value}. Subtype: {subtype}"
                )
        elif set_missing:
            value = None
            if data_type == HASH:
                value = {}
            elif data_type == QUEUE:
                value = deque()
            elif data_type == SET:
                value = set()
            elif data_type == KV:
                value = ""
            self._kv[key] = Value(data_type, value)



"""
Utility functions
"""
from functools import wraps
from ..types import unicode


def encode(characters, encoding: str = "utf-8"):
    """
    Encodes a string with the set encoding defaulted to utf-8
    :param characters: string to encode
    :param encoding: Encoding to set
    :return:
    """
    if isinstance(characters, unicode):
        return characters.encode(encoding=encoding)
    if isinstance(characters, bytes):
        return characters
    return str(characters).encode(encoding=encoding)


def decode(characters) -> str:
    """
    Decodes a string and returns the correct type
    :param characters: String.
    :return: string
    """
    if isinstance(characters, unicode):
        return characters
    if isinstance(characters, bytes):
        return characters.decode("utf-8")
    return str(characters)


def enforce_datatype(data_type, set_missing=True, subtype=None):
    """
    decorator that enforces a data type on a function.
    :param data_type: Data type to enforce
    :param set_missing: Whether to set a value if the key is missing
    :param subtype: subtype to check. This will be for the value
    :return: wrapped function
    """

    def decorator(meth):
        @wraps(meth)
        def inner(self, key, *args, **kwargs):
            self.check_datatype(data_type, key, set_missing, subtype)
            return meth(self, key, *args, **kwargs)

        return inner

    return decorator



"""
Logger configurations, this uses loguru to handle logs
Reference: https://github.com/Delgan/loguru
"""

import os
import sys
import logging
from loguru import logger

logging.root.setLevel(logging.INFO)

root = logging.getLogger()


def configure_log_sink(log_type: str):
    """
    Configures log sing based on the log type and the enrironment
    @param log_type log type could be either info, error, warn, debug, etc
    @returns the log sink to use
    """
    return (
        f"logs/{log_type}.log" if os.environ.get("ENV") == "development" else sys.stdout
    )


def backtrace() -> bool:
    """Configures backtrace based on the env"""
    return os.environ.get("ENV", "development") == "development"


# info log configurations
logger.add(
    sink=configure_log_sink("info"),
    backtrace=backtrace(),
    colorize=True,
    format="<green>{time}</green> <level>{message}</level>",
    # format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
    enqueue=True,
    level="INFO",
)

# error logs
logger.add(
    sink=configure_log_sink("error"),
    backtrace=backtrace(),
    colorize=True,
    format="<green>{time}</green> <level>{message}</level>",
    # format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
    enqueue=True,
    level="ERROR",
)

# debug logs
logger.add(
    sink=configure_log_sink("debug"),
    backtrace=backtrace(),
    colorize=True,
    format="<green>{time}</green> <level>{message}</level>",
    # format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
    enqueue=True,
    level="DEBUG",
)

# warning logs
logger.add(
    sink=configure_log_sink("warn"),
    backtrace=backtrace(),
    colorize=True,
    format="<green>{time}</green> <level>{message}</level>",
    # format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
    enqueue=True,
    level="WARNING",
)

# critical logs
logger.add(
    sink=configure_log_sink("critical"),
    backtrace=backtrace(),
    colorize=True,
    format="<green>{time}</green> <level>{message}</level>",
    # format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
    enqueue=True,
    level="CRITICAL",
)

# trace logs
logger.add(
    sink=configure_log_sink("trace"),
    backtrace=backtrace(),
    colorize=True,
    format="<green>{time}</green> <level>{message}</level>",
    # format="{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}",
    enqueue=True,
    level="TRACE",
)



"""
Threaded stream server
"""
import socketserver


class ThreadedStreamServer:
    """
    Stream server
    """

    def __init__(self, address, handler):
        self.stream_server = None
        self.address = address
        self.handler = handler

    def serve_forever(self):
        """
        Serves the server forever
        """
        handler = self.handler

        class RequestHandler(socketserver.BaseRequestHandler):
            """
            Request handler
            """

            def handle(self) -> None:
                """
                Handles incoming requests
                :return:
                """
                return handler(self.request, self.client_address)

        class ThreadedServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
            """
            Threaded server
            """

            allow_reuse_port = True

        self.stream_server = ThreadedServer(self.address, RequestHandler)
        self.stream_server.serve_forever()

    def stop(self):
        """
        Stops the running server if available
        """
        if self.stream_server:
            self.stream_server.shutdown()



"""
KVault server that is used by the clients to send commands. The QueueServer has the same implementation of the protocol
handler that clients use to parse and send commands. The queue server uses the protocol handler to serialize &
deserialize the messages
"""
from typing import Dict, Callable, AnyStr, Union, Any, List, Tuple
from dataclasses import dataclass, field
import time
from io import BufferedRWPair
from gevent.pool import Pool
from gevent.server import StreamServer
from kvault.infra.logger import logger
from .exceptions import ClientQuit, Shutdown, CommandError, Error
from .protocol_handler import ProtocolHandler
from .types import basestring, Value, unicode
from .utils.mixins import MetaUtils
from .commands import Commands


@dataclass
class Counter:
    """
    Counter contains the counts tracked by the QueueServer.
    :cvar active_connections is the number of current active connections to the queue server
    :cvar commands_processed is the number of commands that the server has processed
    :cvar command_errors is the count of errors encountered by the server
    :cvar connections is the number of connections to the server
    """

    active_connections: int = 0
    commands_processed: int = 0
    command_errors: int = 0
    connections: int = 0


@dataclass
class ServerInfo:
    """
    Counter contains the Server information
    :cvar host is the host the server will run on
    :cvar port is the port the server will run on
    :cvar max_clients is the maximum number of clients that the server will accept connections from
    """

    host: str = "127.0.0.1"
    port: int = 31337
    max_clients: int = 1024


@dataclass
class ServerState:
    """
    Contains the server state
    :cvar kv_store is the in memory Key Value store
    :cvar schedule contains a list of tuples of scheduled commands
    :cvar expiry
    :cvar expiry_map a key value pair where the key is the expiry time and the value is the value. This contains the
    expired data
    """

    kv_store: Dict[AnyStr, Value] = field(default_factory=dict)
    schedule: List[Tuple[Any, Any]] = field(default_factory=list)
    expiry: List[Tuple[float, Any]] = field(default_factory=list)
    expiry_map: Dict[Any, float] = field(default_factory=dict)


class QueueServer(Commands, MetaUtils):
    """
    Queue Server where server send commands to
    """

    # pylint: disable-next=missing-function-docstring
    def __init__(
        self, host: str = "127.0.0.1", port: int = 31337, max_clients: int = 1024
    ):
        self._server_info = ServerInfo(host=host, port=port, max_clients=max_clients)

        self._pool = Pool(max_clients)
        self._server = StreamServer(
            listener=(self._server_info.host, self._server_info.port),
            handle=self.connection_handler,
            spawn=self._pool,
        )
        self._commands = self.get_commands()
        self._protocol = ProtocolHandler()

        self._server_state = ServerState(
            kv_store={}, schedule=[], expiry=[], expiry_map={}
        )

        self._counter = Counter(
            active_connections=0, commands_processed=0, command_errors=0, connections=0
        )

        super().__init__(
            kv_store=self._server_state.kv_store,
            expiry_map=self._server_state.expiry_map,
            expiry=self._server_state.expiry,
            schedule=self._server_state.schedule,
        )

    def connection_handler(self, conn, address):
        """
        Handles a connection given a connection file like object and address
        :param conn: File like socket object
        :param address: address to handle connection on
        """
        logger.info(f"[{self.name}] Connection received: {address}")
        socket_file = conn.makefile("rwb")
        self._counter.active_connections += 1
        while True:
            try:
                self.request_response(socket_file)
            except EOFError:
                logger.info(f"Client went away: {address}")
                socket_file.close()
                break
            except ClientQuit:
                logger.info(f"Client exited: {address}")
                break
            # pylint: disable-next=broad-exception-caught
            except Exception as exc:
                logger.exception(f"Error processing command: {exc}", exc)
        self._counter.active_connections -= 1

    def request_response(self, socket_file: BufferedRWPair):
        """
        Handles the request from a socket file and responds on the protocol handler
        :param socket_file: File like object
        """
        data = self._protocol.handle_request(socket_file)
        try:
            resp = self.respond(data)
        except Shutdown as exc:
            logger.info(f"[{self.name}] Shutting down...")
            self._protocol.write_response(socket_file=socket_file, data=1)
            raise KeyboardInterrupt from exc
        except ClientQuit:
            self._protocol.write_response(socket_file=socket_file, data=1)
            raise
        except CommandError as cmd_error:
            resp = Error(cmd_error.message)
            self._counter.command_errors += 1
        # pylint: disable-next=broad-exception-caught
        except Exception as err:
            logger.error(f"[{self.name}] Unhanded Exception {err}")
            resp = Error(f"Unhandled server error: {err}")
        else:
            self._counter.commands_processed += 1
        self._protocol.write_response(socket_file=socket_file, data=resp)

    def respond(self, data):
        """
        Responds to a given command with the given data. The data is split into 2 parts, the first part is the command
        the second is the data.
        :param data: data to respond to
        :return: response from callback
        """
        if isinstance(data, str):
            try:
                data = data.split()
            # pylint: disable-next=broad-exception-caught
            except Exception as exc:
                raise CommandError(f"Unrecognized request type {data}") from exc
        if not isinstance(data[0], basestring):
            raise CommandError(
                f"First parameter must be command name. Received {data[0]}"
            )

        command = data[0].upper()
        if command not in self._commands:
            logger.error(f"{self.name} Unrecognized command: {command}")
            raise CommandError(f"Unrecognized command: {command}")

        return self._commands[command](*data[1:])

    def get_commands(self) -> Dict[Union[bytes, str], Callable]:
        """
        Returns a mapping of commands to handlers
        :return: Dictionary of commands to handlers
        """
        return dict(
            (
                # Queue commands
                (b"LPUSH", self.lpush),
                (b"RPUSH", self.rpush),
                (b"LPOP", self.lpop),
                (b"RPOP", self.rpop),
                (b"LREM", self.lrem),
                (b"LLEN", self.llen),
                (b"LINDEX", self.lindex),
                (b"LRANGE", self.lrange),
                (b"LSET", self.lset),
                (b"LTRIM", self.ltrim),
                (b"RPOPLPUSH", self.rpoplpush),
                (b"LFLUSH", self.lflush),
                # K/V commands
                (b"APPEND", self.kv_append),
                (b"DECR", self.kv_decr),
                (b"DECRBY", self.kv_decrby),
                (b"DELETE", self.kv_delete),
                (b"EXISTS", self.kv_exists),
                (b"GET", self.kv_get),
                (b"GETSET", self.kv_getset),
                (b"INCR", self.kv_incr),
                (b"INCRBY", self.kv_incrby),
                (b"MDELETE", self.kv_mdelete),
                (b"MGET", self.kv_mget),
                (b"MPOP", self.kv_mpop),
                (b"MSET", self.kv_mset),
                (b"MSETEX", self.kv_msetex),
                (b"POP", self.kv_pop),
                (b"SET", self.kv_set),
                (b"SETNX", self.kv_setnx),
                (b"SETEX", self.kv_setex),
                (b"LEN", self.kv_len),
                (b"FLUSH", self.kv_flush),
                # Hash commands.
                (b"HDEL", self.hdel),
                (b"HEXISTS", self.hexists),
                (b"HGET", self.hget),
                (b"HGETALL", self.hgetall),
                (b"HINCRBY", self.hincrby),
                (b"HKEYS", self.hkeys),
                (b"HLEN", self.hlen),
                (b"HMGET", self.hmget),
                (b"HMSET", self.hmset),
                (b"HSET", self.hset),
                (b"HSETNX", self.hsetnx),
                (b"HVALS", self.hvals),
                # Set commands.
                (b"SADD", self.sadd),
                (b"SCARD", self.scard),
                (b"SDIFF", self.sdiff),
                (b"SDIFFSTORE", self.sdiffstore),
                (b"SINTER", self.sinter),
                (b"SINTERSTORE", self.sinterstore),
                (b"SISMEMBER", self.sismember),
                (b"SMEMBERS", self.smembers),
                (b"SPOP", self.spop),
                (b"SREM", self.srem),
                (b"SUNION", self.sunion),
                (b"SUNIONSTORE", self.sunionstore),
                # Schedule commands.
                (b"ADD", self.schedule_add),
                (b"READ", self.schedule_read),
                (b"FLUSH_SCHEDULE", self.schedule_flush),
                (b"LENGTH_SCHEDULE", self.schedule_length),
                # Misc.
                (b"EXPIRE", self.expire),
                (b"INFO", self.info),
                (b"FLUSHALL", self.flush_all),
                (b"SAVE", self.save_to_disk),
                (b"RESTORE", self.restore_from_disk),
                (b"MERGE", self.merge_from_disk),
                (b"QUIT", self.client_quit),
                (b"SHUTDOWN", self.shutdown),
            )
        )

    def info(self) -> Dict:
        """
        Retrieves the current information of the server
        :return: dictionary mapping of the server information
        """
        return {
            "active_connections": self._counter.active_connections,
            "commands_processed": self._counter.commands_processed,
            "command_errors": self._counter.command_errors,
            "connections": self._counter.connections,
            "keys": len(self._kv),
            "timestamp": time.time(),
        }

    def flush_all(self):
        """
        Clears the store and scheduled commands
        :return: 1 once flushing is successful
        """
        self.kv_flush()
        self.schedule_flush()
        return 1

    def run(self):
        """
        Runs and starts the server
        """
        self._server.serve_forever()

    def add_command(self, command, callback):
        """
        Adds a command to the list of commands supported by the server
        :param command: command name
        :param callback: callback to handle command
        """
        if isinstance(command, unicode):
            command = command.encode("utf-8")
        self._commands[command] = callback



"""
Types
"""
from collections import namedtuple

# pylint: disable-next=invalid-name
unicode = str
basestring = (bytes, str)

Value = namedtuple("Value", ("data_type", "value"))

KV = 0
HASH = 1
QUEUE = 2
SET = 3



"""
Exception types
"""
from collections import namedtuple


class CommandError(Exception):
    """
    Command error exception raised when there is an error with a command
    """

    def __init__(self, message):
        """
        Creates an instance of a command error with a given message.
        :param message: Message of exception
        """
        self.message = message
        super().__init__(message)


class ClientQuit(Exception):
    """Raised when a client quits a connection to the server"""


class Shutdown(Exception):
    """Raised when a there is a shutdown command"""


class ServerError(Exception):
    """Raised when a there is a server error"""


class ServerDisconnect(ServerError):
    """Raised when a there is a server disconnect from the connection pool"""


class ServerInternalError(ServerError):
    """Raised when a there is an internal server error"""


Error = namedtuple("Error", ("message",))



# pylint: disable-next=too-many-lines
"""
Contains all commands performed by the key store
"""
from collections import deque
from typing import Dict, Optional, AnyStr, List, Any, Union
import heapq
import time
import os
import pickle
import datetime
from ..utils.mixins import Guards
from ..exceptions import CommandError, ClientQuit, Shutdown
from ..types import Value, KV, HASH, QUEUE, SET
from ..utils import enforce_datatype, decode


# pylint: disable-next=too-many-public-methods
class Commands(Guards):
    """
    Commands class contains all the commands performed by the key store
    """

    def __init__(
        self,
        kv_store: Optional[Dict[AnyStr, Value]],
        expiry_map: Dict,
        expiry: List,
        schedule: List,
    ):
        """Creates an instance of commands"""
        self._kv: Dict[AnyStr, Value] = kv_store
        self._expiry_map = expiry_map
        self._expiry = expiry
        self._schedule = schedule

        super().__init__(self._kv, self._expiry_map)

    def expire(self, key, nseconds: Union[float, int]):
        """Sets an expiry time for a key in nano-seconds."""
        eta = time.time() + nseconds
        self._expiry_map[key] = eta
        heapq.heappush(self._expiry, (eta, key))

    def clean_expired(self, timestamp=None):
        """
        Performs cleanup of the expired keys
        :param timestamp: timestamp to check against expired keys
        :return: Number of cleanups performed
        """
        _timestamp = timestamp or time.time()
        cleanup_count = 0
        while self._expiry:
            expires, key = heapq.heappop(self._expiry)
            if expires > _timestamp:
                heapq.heappush(self._expiry, (expires, key))
                break

            if self._expiry_map.get(key) == expires:
                del self._expiry_map[key]
                del self._kv[key]
                cleanup_count += 1
        return cleanup_count

    ## Queue commands
    @enforce_datatype(QUEUE)
    def lpush(self, key, *values) -> int:
        """
        Pushes a key with values to a list. This is added to the left of the list
        :param key: identifying Key for the values.
        :param values: Values
        :return: length of values to add
        """
        self._kv[key].value.extendleft(values)
        return len(values)

    @enforce_datatype(QUEUE)
    def rpush(self, key, *values) -> int:
        """
        Pushes a list of values given a key to a list
        :param key: Key to use to identify the values.
        :param values: values to push
        :return: Length of values added
        """
        self._kv[key].value.extend(values)
        return len(values)

    @enforce_datatype(QUEUE)
    def lpop(self, key) -> Any:
        """
        Pops the left item from the queue given a key.
        :param key: Key to pop from the list
        :return: the value of the key
        :raises CommandError if key is not found
        """
        try:
            return self._kv[key].value.popleft()
        except KeyError as key_error:
            raise CommandError(
                f"Failed to find key with error {key_error}. Key {key} does not exist"
            ) from key_error

    @enforce_datatype(QUEUE)
    def rpop(self, key) -> Any:
        """
        Pops the right item from the queue
        :param key: key to pop
        :return: value of key
        :raises CommandError if key is not found
        """
        try:
            return self._kv[key].value.pop()
        except KeyError as error:
            raise CommandError(
                f"Failed to find key with error {error}. Key {key} does not exist"
            ) from error

    @enforce_datatype(QUEUE)
    def lrem(self, key, value) -> int:
        """
        Removes the first occurrence of a value of a given key.
        :param key: Key to use to remove instances of value.
        :param value: Value to remove.
        :return: 0 if value is not found, 1 if value has been removed
        """
        try:
            self._kv[key].value.remove(value)
        except ValueError:
            return 0
        return 1

    @enforce_datatype(QUEUE)
    def llen(self, key) -> int:
        """
        Returns the length of the values of the given key.
        :param key: Key to check for
        :return: length of the values of the given key
        """
        return len(self._kv[key].value)

    @enforce_datatype(QUEUE)
    def lindex(self, key, idx):
        """
        Retrieves the value of a key given a particular index. If the key does not exist, a CommandError is raised
        :param key: Key to check for.
        :param idx: Index of the value to look for in the given key-value pair
        :return: Value at the given index
        :raise: CommandError
        """
        try:
            return self._kv[key].value[idx]
        except KeyError as error:
            raise CommandError(
                f"Failed to find key with error {error}. Key {key} does not exist"
            ) from error

    @enforce_datatype(QUEUE)
    def lset(self, key, idx, value):
        """
        Sets the key, index and value
        :param key: Key to set
        :param idx: Position or index to set the value.
        :param value: The value to set
        :return: 0 if the key does not exist, 1 if the operation is successful
        """
        try:
            self._kv[key].value[idx] = value
        except KeyError:
            return 0
        return 1

    @enforce_datatype(QUEUE)
    def ltrim(self, key, start, stop):
        """
        Trims the key from a given start index to a given stop index(not inclusive). This returns the new trimmed length
        of the queue
        :param key: Key to trim
        :param start: Start of the range
        :param stop: End of the range, exclusive
        :return: length of the trimmed queue
        :raise CommandError if key does not exist
        """
        try:
            trimmed = list(self._kv[key].value)[start:stop]
            self._kv[key] = Value(QUEUE, deque(trimmed))
            return len(trimmed)
        except KeyError as error:
            raise CommandError(
                f"Failed to find key with error {error}. Key {key} does not exist"
            ) from error

    @enforce_datatype(QUEUE)
    def rpoplpush(self, src, dest):
        """
        Pops a given key 'src's value and appends it to the left of the key 'dest' value.
        :param src: Source key whose right value will be popped
        :param dest: Where the popped value will be added
        :return: 0 if the operation fails. For example if either one of the keys does not exist. 1 if the operation
        succeeds
        """
        self.check_datatype(QUEUE, dest, set_missing=True)
        try:
            self._kv[dest].value.appendleft(self._kv[src].value.pop())
        except KeyError:
            return 0
        return 1

    @enforce_datatype(QUEUE)
    def lrange(self, key, start, end=None):
        """
        Returns a range of values of a given key from the start to end. Note that the end value is exclusive
        :param key: Key to get range
        :param start: Start of range
        :param end: End of range(exclusive)
        :return: values in the given range
        :raises: CommandError if key does not exist
        """
        try:
            return list(self._kv[key].value)[start:end]
        except KeyError as error:
            raise CommandError(
                f"Failed to find key with error {error}. Key {key} does not exist"
            ) from error

    @enforce_datatype(QUEUE)
    def lflush(self, key):
        """
        Clears the value of a given key if the key exists
        :param key: Key whose value is to be cleared
        :return: The length of the cleared value
        :raises CommandError if the key does not exist
        """
        try:
            qlen = len(self._kv[key].value)
            self._kv[key].value.clear()
            return qlen
        except KeyError as error:
            raise CommandError(
                f"Failed to find key with error {error}. Key {key} does not exist"
            ) from error

    ## Hash commands
    @enforce_datatype(HASH)
    def hdel(self, key, field) -> int:
        """
        Deletes a field with a given key if available. If the key does not exist a CommandError is raised
        :param key: Key to use.
        :param field: Field to delete
        :return: 1 if the key is successfully deleted. 0 if there is a failure deleting the key
        """
        try:
            value = self._kv[key].value
            if field in value:
                del value[field]
                return 1
            return 0
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hexists(self, key, field) -> int:
        """
        Checks if a field exists in a given key.
        :param key: Key to use
        :param field: value to check for
        :return: 1 if field exists in the key's value, else 0
        :raises: CommandError if the key does not exist
        """
        try:
            value = self._kv[key].value
            return 1 if field in value else 0
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hget(self, key, field) -> Any:
        """
        Gets the value of a field given the key.
        :param key: Key to retrieve value
        :param field: Field to obtain value
        :return: Field's value
        """
        try:
            return self._kv[key].value.get(field)
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hgetall(self, key) -> Value:
        """
        Gets all values of a given key
        :param key: Key to use
        :return: value of a given key if the key exists
        """
        try:
            return self._kv[key].value
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hincrby(self, key, field, incr=1) -> Value:
        """
        Increments a given field in a key by 1
        :param key: Key to use to obtain field.
        :param field: Field to increment
        :param incr: Amount to increment field by. defaulted to 1
        :return: new value
        :raises CommandError if key does not exist
        """
        try:
            self._kv[key].value.setdefault(field, 0)
            self._kv[key].value[field] += incr
            return self._kv[key].value[field]
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hkeys(self, key) -> List[Value]:
        """
        Retrieves values of a given key.
        :param key: Key.
        :return: List of values.
        """
        try:
            return list(self._kv[key].value)
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hlen(self, key) -> int:
        """
        Retrieves the length of the key's values.
        :param key: Key.
        :return: length of keys values
        """
        try:
            return len(self._kv[key].value)
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hmget(self, key, *fields) -> Dict:
        """
        Gets all fields from a key.
        :param key: Key to retrieve values
        :param fields: Fields to extract from values
        :return: dictionary mapping of fields to their values
        """
        try:
            accum = {}
            value = self._kv[key].value
            for field in fields:
                accum[field] = value.get(field)
            return accum
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hmset(self, key, data) -> int:
        """
        Updates key's value with new data
        :param key: Key to update
        :param data: New data to update
        :return: length of data
        """
        try:
            self._kv[key].value.update(data)
            return len(data)
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hset(self, key, field, value) -> int:
        """
        Update the value of a field for a given key.
        :param key: Key to update
        :param field: field to update
        :param value: new value to update field to
        :return: 1 if successful
        """
        try:
            self._kv[key].value[field] = value
            return 1
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hsetnx(self, key, field, value) -> int:
        """
        Updates a key's field with the given value if it is not in the key's value. Will return 1 if the field is not in
        the key's value and update is successful. Otherwise, returns 0 indicating that the field is not in the key's
        value
        :param key: Key to use to retrieve value
        :param field: field to update in the key's value.
        :param value: new value to update field to
        :return: 1 if successful, 0 otherwise
        """
        try:
            kval = self._kv[key].value
            if field not in kval:
                kval[field] = value
                return 1
            return 0
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    @enforce_datatype(HASH)
    def hvals(self, key) -> List:
        """
        Retrieves the values of a given key and returns them as a list of field value pairs
        :param key: Key to retrieve values from
        :return: List of values
        """
        try:
            return list(self._kv[key].value.values())
        except KeyError as error:
            raise CommandError(f"key {key} does not exist. Error: {error}") from error

    # ==== KV Commands
    def unexpire(self, key):
        """
        Removes an expired key from expiry map
        :param key: Key to remove
        """
        self._expiry_map.pop(key, None)

    def _kv_incr(self, key, delta: int) -> Union[Value, Any]:
        """
        Increments the key's value by n. If the key does not exist a new key value is created and they key is returned
        :param key: Key
        :param delta: Delta to increase the value by
        """
        if key in self._kv:
            value = self._kv[key].value + delta
        else:
            value = delta
        self._kv[key] = Value(KV, value)
        return value

    def kv_append(self, key, value):
        """
        Appends a new key value pair
        :param key: Key to add
        :param value: value to add
        """
        if key not in self._kv:
            self.kv_set(key=key, value=value)
        else:
            kv_val = self._kv[key]
            if kv_val.data_type == QUEUE:
                if isinstance(value, list):
                    kv_val.value.extend(value)
                else:
                    kv_val.value.append(value)
            else:
                try:
                    kv_val = Value(kv_val.data_type, kv_val.value + value)
                    self._kv[key] = kv_val
                except Exception as error:
                    raise CommandError(f"Incompatible data-types {value}") from error
        return self._kv[key].value

    def kv_set(self, key, value) -> int:
        """
        Sets a new key value pair and returns 1 if successful
        :param key: Key to add.
        :param value: Value to add
        """
        if isinstance(value, dict):
            data_type = HASH
        elif isinstance(value, list):
            data_type = QUEUE
            value = deque(value)
        elif isinstance(value, set):
            data_type = SET
        else:
            data_type = KV
        self.unexpire(key)
        self._kv[key] = Value(data_type, value)
        return 1

    @enforce_datatype(KV, set_missing=False, subtype=(float, int))
    def kv_decr(self, key):
        """
        Decrements a key's value by 1
        :param key: key to decrement.
        """
        return self._kv_incr(key=key, delta=-1)

    @enforce_datatype(KV, set_missing=False, subtype=(float, int))
    def kv_decrby(self, key, delta: Union[float, int]):
        """
        Decrements a key's value by a set amount n
        :param key: key to decrement
        :param delta: Value to decrement key by
        """
        return self._kv_incr(key=key, delta=-1 * delta)

    def kv_delete(self, key) -> int:
        """
        Deletes a key. returns 1 if successful, 0 if key can not be found
        :param key: Key to delete
        """
        if key in self._kv:
            del self._kv[key]
            return 1
        return 0

    def kv_exists(self, key) -> int:
        """
        Checks if a key exists and has not expired
        :param key: Key to check
        :return: 1 if key is available, 0 if key is not available or has expired
        """
        return 1 if key in self._kv and not self.check_expired(key) else 0

    def kv_get(self, key) -> Optional[Value]:
        """
        Gets the value of a key if the key exists and has not expired
        :param key: Key to check
        :return: Key's value if it exists else None is returned if the key does not exist and has expired
        """
        if key in self._kv and not self.check_expired(key):
            return self._kv[key].value
        return None

    def kv_getset(self, key, value) -> Optional[Value]:
        """
        Retrieves the original value of a key if it exists and has not expired and updates it to a new value and returns
        the old value
        :param key: Key to update
        :param value: new value to update to
        :return: original value of the key
        """
        original_value = None
        if key in self._kv and not self.check_expired(key):
            original_value = self._kv[key].value

        self._kv[key] = Value(KV, value)
        return original_value

    @enforce_datatype(KV, set_missing=False, subtype=(float, int))
    def kv_incr(self, key):
        """
        Increments the value of a key by 1
        :param key: key to increment value by
        """
        return self._kv_incr(key, 1)

    @enforce_datatype(KV, set_missing=False, subtype=(float, int))
    def kv_incrby(self, key, delta):
        """
        Increments the value of a key by n
        :param key: key to increment
        :param delta: delta to increment the value by
        """
        return self._kv_incr(key, delta)

    def kv_mdelete(self, *keys):
        """
        Tries deleting keys if available. Returns the total number of keys deleted
        :param keys: Keys to delete
        :return: number of keys deleted
        """
        deleted_key_count = 0
        for key in keys:
            try:
                del self._kv[key]
            except KeyError:
                pass
            else:
                deleted_key_count += 1
        return deleted_key_count

    def kv_mget(self, *keys) -> List[Optional[Value]]:
        """
        Retrieves the values of keys and returns a list of values for the keys. In this list, some values will be None.
        indicating that a value for that key was not found. For example: [1, None, 3, 5, None] for keys
        ['a', 'b', 'c', 'd', 'e']. This means values for keys 'b' and 'e' could not be found because they either do not
        exist or the keys have expired.
        :param keys: Keys
        :return: list of values
        """
        accum = []
        for key in keys:
            if key in self._kv and not self.check_expired(key):
                accum.append(self._kv[key].value)
            else:
                accum.append(None)
        return accum

    def kv_mpop(self, *keys) -> List[Optional[Value]]:
        """
        Retrieves the values of keys and removes them. This returns a list of values for the keys. In this list,
        some values will be None. indicating that a value for that key was not found. For example: [1, None, 3, 5, None]
        for keys  ['a', 'b', 'c', 'd', 'e']. This means values for keys 'b' and 'e' could not be found because they
        either do not exist or the keys have expired.

        :param keys: Keys
        :return: list of values
        """
        accum = []
        for key in keys:
            if key in self._kv and not self.check_expired(key):
                accum.append(self._kv.pop(key).value)
            else:
                accum.append(None)
        return accum

    def kv_mset(self, __data: Optional[Dict] = None, **kwargs) -> int:
        """
        Adds and unexpires key value pairs __data along with kwargs. This returns the number of updates performed.
        :param __data: key value pairs to add
        :param kwargs: Additional key value pairs to add
        :return: number of updates performed
        """
        update_count = 0
        data = {}
        if __data is not None:
            data.update(__data)
        if kwargs is not None:
            data.update(kwargs)

        for key_, _ in data.items():
            self.unexpire(key_)
            self._kv[key_] = Value(KV, data[key_])
            update_count += 1
        return update_count

    def kv_msetex(self, data: Dict, expires: Union[float, int]):
        """
        Calls on kv_mset with the provided key value pairs and sets an expiry time for the keys
        :param data: Key value pairs to add.
        :param expires: when a key should expire. This will apply to all keys in data
        """
        self.kv_mset(data)
        for key in data:
            self.expire(key, expires)

    def kv_pop(self, key) -> Optional[Value]:
        """
        Pops a key if the key is available and if it has expired
        :param key: Key to pop.
        :return: The value of the popped key
        """
        if key in self._kv and not self.check_expired(key):
            return self._kv.pop(key).value
        return None

    def kv_setnx(self, key, value) -> int:
        """
        Sets a value of a key if it's not available and has not expired.
        :param key: Key to add
        :param value: Value of key
        :return: 1 if the key is added, 0 if no operation is performed
        """
        if key in self._kv and not self.check_expired(key):
            return 0
        self.unexpire(key)
        self._kv[key] = Value(KV, value)
        return 1

    def kv_setex(self, key: Any, value: Value, expires: Union[float, int]) -> int:
        """
        Sets the value of a key and the expiry of the key.
        :param key: Key to set
        :param value: Value of the key
        :param expires: Expiry time of the key
        """
        self.kv_set(key, value)
        self.expire(key, expires)
        return 1

    def kv_len(self) -> int:
        """
        Returns the length of the keys
        :return: length of the keys.
        """
        return len(self._kv)

    def kv_flush(self) -> int:
        """
        Clears the keys, expiry and expiry_map and returns the original length of the keys.
        :return: length of the original key
        """
        kvlen = self.kv_len()
        self._kv.clear()
        self._expiry = []
        self._expiry_map = {}
        return kvlen

    # ====== Set Commands
    @enforce_datatype(SET)
    def sadd(self, key, *members):
        """
        Updates the key with the members
        :param key: Key to update
        :param members: members to add.
        :return: The new length of the key's values
        """
        try:
            self._kv[key].value.update(members)
            return len(self._kv[key].value)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def scard(self, key):
        """
        Returns the length of the key's values for a set
        :param key: Key to use
        :return: length of new key
        """
        return len(self._kv[key].value)

    @enforce_datatype(SET)
    def sdiff(self, key, *keys) -> List:
        """
        Returns the difference between the key and the set of keys
        :param key: Key to use
        :param keys: set of keys to check.
        :return: list of the difference between the keys.
        """
        try:
            src = set(self._kv[key].value)
            for key_ in keys:
                self.check_datatype(SET, key_)
                src -= self._kv[key_].value
            return list(src)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def sdiffstore(self, dest, key, *keys) -> int:
        """
        Finds the difference between the key and as set of keys before setting the value of the key "dest" with the
        difference from the set.
        :param dest: Destination key to store values
        :param key: Key to check for difference
        :param keys: collection of keys to check for difference
        :return: length of difference between the keys
        """
        try:
            src = set(self._kv[key].value)
            for key_ in keys:
                self.check_datatype(SET, key_)
                src -= self._kv[key_].value
            self.check_datatype(SET, dest)
            self._kv[dest] = Value(SET, src)
            return len(src)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def sinter(self, key, *keys) -> List:
        """
        Checks the intersection between key and set of keys and returns the list
        :param key: Key
        :param keys: set of keys to check
        :return: list of intersection
        """
        try:
            src = set(self._kv[key].value)
            for key_ in keys:
                self.check_datatype(SET, key_)
                src &= self._kv[key_].value
            return list(src)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def sinterstore(self, dest, key, *keys) -> int:
        """
        Checks the intersection between key and keys and stores the intersection in "dest"
        :param dest: Destination key
        :param key: Source key
        :param keys: Set of keys
        :return: length of intersection
        """
        try:
            src = set(self._kv[key].value)
            for key_ in keys:
                self.check_datatype(SET, key_)
                src &= self._kv[key_].value
            self.check_datatype(SET, dest)
            self._kv[dest] = Value(SET, src)
            return len(src)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def sismember(self, key, member):
        """
        Checks if a member is a member of a key's value.
        :param key: Key to check
        :param member: Member to check
        :return: 1 if member is a member of a set, else returns 0
        """
        try:
            return 1 if member in self._kv[key].value else 0
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def smembers(self, key):
        """
        returns the members of a key.
        :param key: Key to use
        :return: member values of the key.
        """
        try:
            return self._kv[key].value
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def spop(self, key, number_to_pop=1) -> List[Value]:
        """
        Pops n values from the key and returns them
        :param key: Key to pop
        :param number_to_pop: number of values
        :return: accumulated values
        """
        accum = []
        for _ in range(number_to_pop):
            try:
                accum.append(self._kv[key].value.pop())
            except KeyError:
                break
        return accum

    @enforce_datatype(SET)
    def srem(self, key, *members) -> int:
        """
        Removes a set of members from the given key's values
        :param key: Key to remove
        :param members: set of members
        :return: number of members removed
        """
        remove_count = 0
        for member in members:
            try:
                self._kv[key].value.remove(member)
            except KeyError:
                pass
            else:
                remove_count += 1
        return remove_count

    @enforce_datatype(SET)
    def sunion(self, key, *keys) -> List:
        """
        Returns the union of key and keys
        :param key: Key to check for union
        :param keys: Keys to check for union
        :return: list of union pairs
        """
        try:
            src = set(self._kv[key].value)
            for key_ in keys:
                self.check_datatype(SET, key_)
                src |= self._kv[key_].value
            return list(src)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    @enforce_datatype(SET)
    def sunionstore(self, dest, key, *keys) -> int:
        """
        Gets the union of key and keys and stores it in dest
        :param dest: Destination key
        :param key: source key
        :param keys: Set or collection of keys
        :return: length of union
        """
        try:
            src = set(self._kv[key].value)
            for key_ in keys:
                self.check_datatype(SET, key_)
                src |= self._kv[key_].value
            self.check_datatype(SET, dest)
            self._kv[dest] = Value(SET, src)
            return len(src)
        except KeyError as error:
            raise CommandError(f"Key {key} does not exist {error}") from error

    def _get_state(self) -> Dict[str, Any]:
        """
        Returns the current state of the store
        :return: dictionary mapping of state keys to mappings
        """
        return {"kv": self._kv, "schedule": self._schedule}

    def _set_state(self, state: Dict[str, Any], merge=False):
        """
        Sets the state to a new given state
        :param state: New state
        :param merge: whether to merge the state with the new state
        """
        if not merge:
            self._kv = state["kv"]
            self._schedule = state["schedule"]
        else:

            def merge(orig, updates):
                orig.update(updates)
                return orig

            self._kv = merge(state["kv"], self._kv)
            self._schedule = state["schedule"]

    def save_to_disk(self, filename) -> bool:
        """Saves the current state to disk given a filename."""
        with open(filename, "wb") as file_handle:
            pickle.dump(self._get_state(), file_handle, pickle.HIGHEST_PROTOCOL)
        return True

    def restore_from_disk(self, filename: str, merge=False):
        """
        Restores a file from disk if the file exists. It also performs a merge of the stored state with the current
        state if the merge argument is set to True
        :param filename: filename to use
        :param merge: whether to merge to the new state
        """
        if not os.path.exists(filename):
            return False
        with open(filename, "rb") as file_handle:
            state = pickle.load(file_handle)
        self._set_state(state, merge=merge)
        return True

    def merge_from_disk(self, filename) -> bool:
        """Merges stored file from disk with the current state of store"""
        return self.restore_from_disk(filename, merge=True)

    def client_quit(self):
        """Raises a client quit exception"""
        raise ClientQuit("client closed connection")

    def shutdown(self):
        """Raises a shutdown exception"""
        raise Shutdown("shutting down")

    def _decode_timestamp(self, timestamp: str) -> datetime:
        """Decodes a given timestamp"""
        timestamp_ = decode(timestamp)
        fmt = "%Y-%m-%d %H:%M:%S"
        if "." in timestamp_:
            fmt = fmt + ".%f"
        try:
            return datetime.datetime.strptime(timestamp_, fmt)
        except ValueError as error:
            raise CommandError(
                f"Timestamp {timestamp} must be formatted Y-m-d H:M:S"
            ) from error

    def schedule_add(self, timestamp, data) -> int:
        """
        Adds a new schedule to recorded schedules
        :param timestamp: timestamp to decode
        :param data: data to add to schedule
        """
        try:
            decoded_timestamp = self._decode_timestamp(timestamp)
            heapq.heappush(self._schedule, (decoded_timestamp, data))
            return 1
        except Exception as error:
            raise CommandError(
                f"timestamp {timestamp} does not exist {error}"
            ) from error

    def schedule_read(self, timestamp=None) -> List:
        """
        Reads a given timestamp and returns the schedules that are less than the given timestamp. This removes the
        schedules from the given schedule record.
        """
        try:
            decoded_timestamp = self._decode_timestamp(timestamp)
            accum = []
            while self._schedule and self._schedule[0][0] <= decoded_timestamp:
                _, data = heapq.heappop(self._schedule)
                accum.append(data)
            return accum
        except Exception as error:
            raise CommandError(f"Failed to read {timestamp}. Error: {error}") from error

    def schedule_flush(self) -> int:
        """
        Flushes the schedule and returns the previous length.
        """
        schedule_len = self.schedule_length()
        self._schedule = []
        return schedule_len

    def schedule_length(self) -> int:
        """
        Returns the length of the schedule
        """
        return len(self._schedule)
